import os
import pytest
import lpsds.handlers

def fake_make_dir(path):
    return path

@pytest.fixture(autouse=True)
def set_mocks(monkeypatch):
    monkeypatch.setattr(os, "makedirs", fake_make_dir)

class TestGetFullPath():
    @pytest.fixture
    def bucket(self):
        return '/bucket'
    
    @pytest.fixture
    def fh_obj(self, bucket):
        return lpsds.handlers.LocalFileHandler(bucket=bucket)

    def test_full_parameters_setting(self, bucket, fh_obj):
        dir_name = 'directory'
        prefix = 'prefix'
        fname = 'filename.ext'
        
        ret = fh_obj.get_full_path(dir_name=dir_name, prefix=prefix, fname=fname)
        assert ret == os.path.join(bucket, dir_name, prefix, fname)

    def test_full_path_setting(self, bucket, fh_obj):
        fname = 'directory/prefix/filename.ext'
        ret = fh_obj.get_full_path(fname=fname)
        assert ret == os.path.join(bucket, fname)
 


import collections.abc
import numpy as np
import pandas as pd
import sklearn.pipeline
from typing import Union
from .metrics import bootstrap_estimate

class ObjectView(dict):
    """Data structure that improves a regular map structure so
    its components can be accessed as map['value'] or map.value.
    It works only for string keys.
    """
    def __getattr__(self, name):
        if name in self: return self[name]
        else: raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self: del self[name]
        else: raise AttributeError("No such attribute: " + name)


def to_list(val):
    """
    If val is an iterable collection other than a string, returns val,
    otherwise, returs [val].
    """
    if isinstance(val, collections.abc.Iterable) and type(val) is not str:
        return val
    else: return [val]


def smart_tuple(vec):
    """
    If len(vec) == 1, returns vec[0], otherwise, returns vec as a tuple.
    """
    if len(vec) == 1: return vec[0]
    else: return tuple(vec)


def keep(df, index=None, columns=None, inplace=True):
    """
    def keep(index=None, columns=None, inplace=True)
    
    The opposite of pd.drop function. In this function, you specify the index / columns you
    want to keep. I.e. the function will drop everything that does not match the index/cools
    specified.
    
    IMPORTANT REMARKS:
     - differently than the drop function, "inplace" optioon defaults to True.
     - index and columns are FILTER, not the final list of values to be kept.
    
    Returns: the filtered dataframe if inplace=True or None otherwise.
    """
    
    if index is not None:
        index = df.loc[~index].index
    
    if columns is not None:
        columns = df.columns[~columns]
    
    return df.drop(index=index, columns=columns, inplace=inplace)


def confusion_matrix_annotation(cm: np.array, fmt_str: str='${mean:.2f}^{{+{e_max:.2f}}}_{{-{e_min:.2f}}}$', use_difference: bool=True, seed: int=None) -> pd.DataFrame:
    """
    def confusion_matrix_annotation(cm: np.array, fmt_str: str='${mean:.2f}^{{+{e_max:.2f}}}_{{-{e_min:.2f}}}$', 
                                    use_difference: bool=True, seed: int=None) -> pd.DataFrame
    
    Creates a pandas dataframe with fancy annotation to be used in seaborn.heatmap. It automatically calculates
    mean and error margins using lpsds.metrics.bootstrap_estimate.

    Input parameters:
      - cm: a numpy.array object with the confusion matrix organized as [num_folds, num_rows, num_cols].
      - fmt_str: a tring defining how the numbers should be presented. Defaults to mean^{+err max}_{-err_min}.
                 you must follow the right name convention if you want to use your own fmt string. Use the convention as:
                    - mean: the mean value
                    - e_min: the lower threshold error margin
                    - e_max: the higher threshold error margin.
      - use_difference: if True (default), will present e_min as (mean-e_min) and e_max as (e_max - mean).
      - seed: random seed.

    Returns: a dataframe with shape [num_rows, num_cols] containing in each cell the defined annotation.

    ATTENTION: when passing the annotation dataframe to seaborn.heatmap, you MUST set fmt=''. Example:
               seaborn.heatmap(annot=my_annotation_df, fmt='') 
    """
    
    num_rows, num_cols = cm.shape[1:]
    ret = pd.DataFrame(np.empty([num_rows, num_cols], dtype=str))
    for r in range(num_rows):
        for c in range(num_cols):
            mean, e_min, e_max = bootstrap_estimate(cm[:,r,c], seed=seed)
            if use_difference:
                e_min = mean - e_min
                e_max -= mean
            ret.iloc[r][c] = fmt_str.format(mean=mean, e_min=e_min, e_max=e_max)
    
    return ret



def pretty_title(text :str, abbrev_threshold :int=4) -> str:
    """
    def pretty_title(text :str, abbrev_threshold :int=4) -> str:

        Returns a pretty string. If it is an abbreviation (text > than abbrev_threshold),
        the ext will be put in upper case. Otherwise, it will have all its first letters
        in upper case the rest in lower case.
    """
    return text.upper() if len(text) <= abbrev_threshold else text.capitalize()



def error_bar_roc(df :pd.DataFrame, tp_col_name :str='true_positive', fp_col_name: str='false_positive', 
                    fold_col_name :str='fold', threshold_col_name :str='threshold', num_points: int=100) -> pd.DataFrame:
    """
    def error_bar_roc(df :pd.DataFrame, tp_col_name :str='true_positive', fp_col_name: str='false_positive', 
                        fold_col_name :str='fold', threshold_col_name :str='threshold', num_points: int=100) -> pd.DataFrame:

    Creates a new ROC curve for each CV fold, making sure that all folds share the same false alarm values, allowing plotting
    ROC curve with its error marging.

    Input:
      - df: a dataframe containing the ROC values in long format (fold x tpr x fpr x threshold). All mentioned ROC values
            must be present in df.
      - tp_col_name: the name of the column in df holding the ROC true positive values.
      - fp_col_name: the name of the column in df holding the ROC false positive values.
      - fold_col_name: the name of the column in df holding the fold ID of each ROC value.
      - threshold_col_name: the name of the column in df holding the ROC threshold values.
      - num_points: how many points will be used to compute the new ROCS.
    
    Return: a dataframe where all TP values (for all folds) were computed for the same FP values,
            so all ROCS have the save FP values and therefore, the same length (num_points).
            Columns in the new dataframe are::
                - tp_col_name: the name of the column in df holding the new ROC true positive values.
                - fp_col_name: the name of the column in df holding the new ROC false positive values.
                - fold_col_name: the name of the column in df holding the fold ID of each new ROC value.
    """
    #false alarm must be in ascending order otherwise np.interp will not work.
    df = df.sort_values([fold_col_name, threshold_col_name], ascending=False)
    
    num_folds = len(df[fold_col_name].unique())
    mean_fpr = np.linspace(0, 1, num_points)
    
    rocs_list = []
    for fold in range(num_folds):
        fold_roc = df.loc[df[fold_col_name] == fold]
        interp_tpr = np.interp(mean_fpr, fold_roc[fp_col_name], fold_roc[tp_col_name])
        roc_aux = pd.DataFrame({fp_col_name : mean_fpr, tp_col_name : interp_tpr, fold_col_name : fold})
        rocs_list.append(roc_aux)
    
    return pd.concat(rocs_list, ignore_index=True)



def pipeline_split(model :sklearn.pipeline.Pipeline, X: Union[pd.DataFrame, np.ndarray]=None):
    """
    def pipeline_input_split(model :sklearn.pipeline.Pipeline, X: Union[pd.DataFrame, np.ndarray]=None):
    
    Breaks a pipeline into 2 sections: preprocessing and estimator sections. If X is provided, the function
    will apply the pre-processing chain on it, and return it in a stage right before it would be fed to the
    pipeline estimator.

    Input:
      - model: a sklearn-li Pipeline object
      - X: an input dataset.
    
      Returns:
        - A list with the pipeline pre-processing steps.
        - The estimator object at the end of the pipeline.
        - Pre-processed X (the values of X right before it will be fed to the pipeline estimator)
    """
    
    pre_processing = model.steps[:-1]
    estimator = model.steps[-1]
    
    if X is not None:
        for _, func in pre_processing:
            X = func.transform(X)
    
    return pre_processing, estimator, X



import os
import datetime
import numpy as np
import pandas as pd
import pytest
import mlflow
import lpsds.metrics
from lpsds.mlflow import MLFlow
from lpsds.utils import ObjectView

class MLFlowBase:

    class MockedRunClient:
        def get_metric_history(self, run_id, metric_name):

            sp_list = []
            f1_list = []
            for i in range(5):
                for j, v in enumerate([sp_list, f1_list]):
                    w = 1 if j==0 else 10
                    obj = ObjectView()
                    obj.value = w * (i+1)
                    obj.step = i+1
                    v.append(obj)

            ret_map = {}
            for m, v in zip(['sp_mean', 'f1_mean', 'acc'], [0.9, 0.95, 0.99]):
                obj = ObjectView()
                obj.value = v
                obj.step = 0
                ret_map[m] = [ obj ]

            ret_map['sp'] = sp_list
            ret_map['f1'] = f1_list
            return ret_map[metric_name]


    @pytest.fixture
    def mlf_obj(self):
        return MLFlow()

    @staticmethod
    def mocked_active_run():
        """
        Mocks mlflow.active_run behavior
        """
        ret = ObjectView()
        ret.info = ObjectView()
        ret.info.run_id = '112233'
        ret.info.experiment_id = '112233'
        return ret
    
    @staticmethod
    def mocked_get_run(run_id):
        """
        Mocks mlflow.active_run behavior
        """
        ret = ObjectView()
        ret.info = ObjectView()
        ret.info.run_id = run_id
        ret.info.experiment_id = run_id + '987'
        ret.data = ObjectView()
        ret.data.metrics = dict(
            sp_mean = 0.9,
            f1_mean = 0.95,
            acc = 0.99,
            sp = 0,
            f1 = 10,
        )
        ret.data.params = dict(
            param_str ='my string',
            param_int = '123',
            param_float = '3.14',
            param_bool = 'True',
            param_date = '2022-12-24',
        )
        return ret
    
    @staticmethod
    def mocked_get_experiment(exp_id):
        ret = ObjectView()
        ret.name = f'test_experiment_{exp_id}'
        return ret


    def mocked_download_artifacts(run_id, artifact_path, dst_path):
        df = pd.DataFrame({'a' : [1,2,3], 'b' : [11,22,33]})
        fname = os.path.split(artifact_path)[1]
        local_name = os.path.join(dst_path, fname)
        df.to_parquet(local_name)
        return local_name


    @pytest.fixture(autouse=True)
    def set_mlflow_patches(self, monkeypatch, request):
        if 'noautofixt' in request.keywords: return
        monkeypatch.setattr(mlflow, 'active_run', MLFlowBase.mocked_active_run)
        monkeypatch.setattr(mlflow, 'get_run', MLFlowBase.mocked_get_run)
        monkeypatch.setattr(mlflow, 'get_experiment', MLFlowBase.mocked_get_experiment)
        monkeypatch.setattr(mlflow.tracking, 'MlflowClient', MLFlowBase.MockedRunClient)
        monkeypatch.setattr(mlflow.artifacts, 'download_artifacts', MLFlowBase.mocked_download_artifacts)


class TestInit(MLFlowBase):

    def test_call_with_id(self):
        a = MLFlow('12345')
        assert a.run_id == '12345'

    @pytest.mark.noautofixt
    def test_call_without_id_without_experiment(self):
        with pytest. raises(ValueError):
            MLFlow()


    def test_call_without_id(self):
        a = MLFlow()
        assert a.run_id == '112233'


class TestLogStatistics(MLFlowBase):

    @staticmethod
    def mocked_log_metrics(metrics_map):
        """
        Mocks mlflow.log_metrics behavior
        """
        return None

    @staticmethod
    def mocked_bootstrap_estimate(vec, ci=95, n_boot=1000, seed=None):
        """
        Returns a mocked result for bootstrap_estimate
        """
        return vec.mean(), vec.mean()-vec.min(), vec.mean()+vec.max()
    
    @staticmethod
    def mocked_log_metric(key, value, step):
        return None

    @pytest.fixture(autouse=True)
    def set_patches(self, monkeypatch):
        monkeypatch.setattr(lpsds.metrics, 'bootstrap_estimate', TestLogStatistics.mocked_bootstrap_estimate)
        monkeypatch.setattr(mlflow, 'log_metrics', TestLogStatistics.mocked_log_metrics)
        monkeypatch.setattr(mlflow, 'log_metric', TestLogStatistics.mocked_log_metric)

    @pytest.fixture
    def cv_map(self):
        """
        Returns the mocked data for each test
        """        
        return {
            'test_sp' : np.array([1,2,3,4,5]), #mean = 3, err_min=2, err_max=8
            'estimators' : np.array([1,2,3,4,5]),
            'test_auc' : np.array([10,2,30,4,50]), #mean=19.2, err_min=17.2, err_max=69.2
        }

    def test_only_2_returned(self, cv_map, mlf_obj):
        """
        Only 2 metrics must be returned
        """
        ret = mlf_obj.log_statistics(cv_map)
        assert len(ret) == 2


    def test_metric_names(self, cv_map, mlf_obj):
        """
        Only SP and AUC must be returned
        """
        ret = mlf_obj.log_statistics(cv_map)
        assert 'sp' in ret
        assert 'auc' in ret

    def test_metric_values(self, cv_map, mlf_obj):
        """
        Only metric values must be correct
        """
        ret = mlf_obj.log_statistics(cv_map)
        assert ret['sp']['sp_mean'] == 3
        assert ret['sp']['sp_err_min'] == 2
        assert ret['sp']['sp_err_max'] == 8

        assert ret['auc']['auc_mean'] == 19.2
        assert ret['auc']['auc_err_min'] == 17.2
        assert ret['auc']['auc_err_max'] == 69.2
    
    @staticmethod
    def assert_metrics_logged(metric_map):
        assert metric_map['sp_mean'] == 3
        assert metric_map['sp_err_min'] == 2
        assert metric_map['sp_err_max'] == 8

    def test_mlflow_submission(self, monkeypatch, mlf_obj):
        met_map = {'test_sp' : np.array([1,2,3,4,5])} #mean = 3, err_min=2, err_max=8
        monkeypatch.setattr(mlflow, 'log_metrics', TestLogStatistics.assert_metrics_logged)
        mlf_obj.log_statistics(met_map)
    

    def test_metrics_vector_logged(self, monkeypatch, mlf_obj):
        self.metric_list = []
        def mocked_log_metric(key, value, step):
            self.metric_list.append((key, value, step))


        met_map = {'test_sp' : np.array([1,2,3,4,5])} #mean = 3, err_min=2, err_max=8
        monkeypatch.setattr(mlflow, 'log_metric', mocked_log_metric)
        mlf_obj.log_statistics(met_map)
        for i, items in enumerate(self.metric_list):
            k,v,s = items
            assert k == 'sp'
            assert v == i+1
            assert s == i



class TestLogArtifact(MLFlowBase):

    @staticmethod
    def assert_file_exists(temp_file_name, folder):
        assert os.path.exists(temp_file_name)
        assert folder == 'mlflow_test_folder'

    @staticmethod
    def assert_right_content(temp_file_name, folder):
        df = pd.read_parquet(temp_file_name)
        assert df.shape[0] == 3
        assert df.shape[1] == 2

        assert df.a.iloc[0] == 1
        assert df.a.iloc[1] == 2
        assert df.a.iloc[2] == 3

        assert df.b.iloc[0] == 40
        assert df.b.iloc[1] == 50
        assert df.b.iloc[2] == 60


    @pytest.fixture
    def df(self):
        return pd.DataFrame({'a' : [1,2,3], 'b' : [40,50,60]})

    def test_file_exists(self, monkeypatch, df, mlf_obj):
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogArtifact.assert_file_exists)    
        mlf_obj.log_artifact(df, 'filename.parquet', 'mlflow_test_folder', save_func=df.to_parquet, fname_param_name='path', object_param_name=None)


    def test_file_correct(self, monkeypatch, df, mlf_obj):
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogArtifact.assert_right_content)
        mlf_obj.log_artifact(df, 'filename.parquet', 'mlflow_test_folder', save_func=df.to_parquet, fname_param_name='path', object_param_name=None)



class TestGetRunID(MLFlowBase):
    def test_return(self, mlf_obj):
        assert mlf_obj.get_run_id() == '112233'


class TestGetMetrics(MLFlowBase):

    def test_num_returns(self, mlf_obj):
        met = mlf_obj.get_metrics(as_dict=True)
        assert len(met) == 5    

    def test_scalars(self, mlf_obj):
        met = mlf_obj.get_metrics(as_dict=True)
        assert met['sp_mean'] == 0.9
        assert met['f1_mean'] == 0.95
        assert met['acc'] == 0.99


    def test_object_view(self, mlf_obj):
        met = mlf_obj.get_metrics(as_dict=True)
        assert isinstance(met, ObjectView)


    def test_vectors(self, mlf_obj):
        met = mlf_obj.get_metrics(as_dict=True)
        assert np.array_equal(met['sp'], np.array([1,2,3,4,5]))
        assert np.array_equal(met['f1'], np.array([10,20,30,40,50]))


        
    def test_dataframe_structure(self, mlf_obj):
        met = mlf_obj.get_metrics()
        assert met.shape[0] == 13
        assert met.shape[1] == 3
        assert met.columns[0] == 'metric'
        assert met.columns[1] == 'step'
        assert met.columns[2] == 'value'

    

    def test_scalars_dataframe(self, mlf_obj):
        met = mlf_obj.get_metrics()
        assert met.loc[met.metric == 'sp_mean', 'value'].iloc[0] == 0.9
        assert met.loc[met.metric == 'f1_mean', 'value'].iloc[0] == 0.95
        assert met.loc[met.metric == 'acc', 'value'].iloc[0] == 0.99


    def test_vectors_dataframe(self, mlf_obj):
        met = mlf_obj.get_metrics()
        assert np.array_equal(met.loc[met.metric == 'sp', 'value'].to_numpy(), np.array([1,2,3,4,5]))
        assert np.array_equal(met.loc[met.metric == 'f1', 'value'].to_numpy(), np.array([10,20,30,40,50]))




class TestGetArtifact(MLFlowBase):

    def test_operation_no_folder(self, mlf_obj):
        df = mlf_obj.get_artifact('my_df', load_func=pd.read_parquet)
        assert df.shape[0] == 3
        assert df.shape[1] == 2
        assert df.iloc[0].a == 1
        assert df.iloc[0].b == 11
        assert df.iloc[1].a == 2
        assert df.iloc[1].b == 22
        assert df.iloc[2].a == 3
        assert df.iloc[2].b == 33

    def test_operation_with_folder(self, mlf_obj):
        df = mlf_obj.get_artifact('my_df', 'folder/path', load_func=pd.read_parquet)
        assert df.shape[0] == 3
        assert df.shape[1] == 2
        assert df.iloc[0].a == 1
        assert df.iloc[0].b == 11
        assert df.iloc[1].a == 2
        assert df.iloc[1].b == 22
        assert df.iloc[2].a == 3
        assert df.iloc[2].b == 33


class TestGetExperiment(MLFlowBase):
    def test_positive_case(self):
        a = MLFlow('123')
        assert a.get_experiment().name == 'test_experiment_123987'


class TestGetParams(MLFlowBase):

    def test_num_returns(self, mlf_obj):
        par = mlf_obj.get_params()
        assert len(par) == 5

    def test_str_return(self, mlf_obj):
        par = mlf_obj.get_params(infer_types=False)
        assert par['param_str'] == 'my string'
        assert par['param_int'] == '123'
        assert par['param_float'] == '3.14'
        assert par['param_bool'] == 'True'
        assert par['param_date'] == '2022-12-24'


    def test_infer_type_return(self, mlf_obj):
        par = mlf_obj.get_params(infer_types=True)
        assert par['param_str'] == 'my string'
        assert par['param_int'] == 123
        assert par['param_float'] == 3.14
        assert par['param_bool'] == True
        assert par['param_date'] == datetime.date(2022, 12, 24)



    def test_object_view(self, mlf_obj):
        par = mlf_obj.get_params()
        assert isinstance(par, ObjectView)


class TestLogDataFrame(MLFlowBase):
    @pytest.fixture
    def df(self):
        return pd.DataFrame({'a' : [1,2,3], 'b' : [40,50,60]})

    def test_file_exists(self, monkeypatch, df, mlf_obj):
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogArtifact.assert_file_exists)    
        mlf_obj.log_dataframe(df, 'filename', 'mlflow_test_folder')


    def test_file_correct(self, monkeypatch, df, mlf_obj):
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogArtifact.assert_right_content)
        mlf_obj.log_dataframe(df, 'filename', 'mlflow_test_folder')


class TestLogNumpy(MLFlowBase):
    @staticmethod
    def assert_file_exists(temp_file_name, folder):
        assert os.path.exists(temp_file_name)
        assert folder == 'mlflow_test_folder'

    @staticmethod
    def assert_right_content(temp_file_name, folder):
        df = np.load(temp_file_name)
        assert df.shape[0] == 2
        assert df.shape[1] == 3

        assert df[0][0] == 1
        assert df[0][1] == 2
        assert df[0][2] == 3

        assert df[1][0] == 40
        assert df[1][1] == 50
        assert df[1][2] == 60


    @pytest.fixture
    def df(self):
        return np.array([[1,2,3],[40,50,60]])

    def test_file_exists(self, monkeypatch, df, mlf_obj):
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogNumpy.assert_file_exists)    
        mlf_obj.log_numpy(df, 'filename', 'mlflow_test_folder')


    def test_file_correct(self, monkeypatch, df, mlf_obj):
        monkeypatch.setattr(mlflow, 'log_artifact', TestLogNumpy.assert_right_content)
        mlf_obj.log_numpy(df, 'filename', 'mlflow_test_folder')



class TestGetDataFrame(MLFlowBase):

    def test_operation_no_folder(self, mlf_obj):
        df = mlf_obj.get_dataframe('my_df')
        assert df.shape[0] == 3
        assert df.shape[1] == 2
        assert df.iloc[0].a == 1
        assert df.iloc[0].b == 11
        assert df.iloc[1].a == 2
        assert df.iloc[1].b == 22
        assert df.iloc[2].a == 3
        assert df.iloc[2].b == 33

    def test_operation_with_folder(self, mlf_obj):
        df = mlf_obj.get_dataframe('my_df', 'folder/path')
        assert df.shape[0] == 3
        assert df.shape[1] == 2
        assert df.iloc[0].a == 1
        assert df.iloc[0].b == 11
        assert df.iloc[1].a == 2
        assert df.iloc[1].b == 22
        assert df.iloc[2].a == 3
        assert df.iloc[2].b == 33




class TestGetNumpy(MLFlowBase):

    @staticmethod
    def mocked_download_artifacts_numpy(run_id, artifact_path, dst_path):
        df = np.array([[1,2,3],[11,22,33]])
        fname = os.path.split(artifact_path)[1]
        local_name = os.path.join(dst_path, fname + '.npy')
        np.save(local_name, df)
        return local_name


    def test_operation_no_folder(self, mlf_obj, monkeypatch):
        monkeypatch.setattr(mlflow.artifacts, 'download_artifacts', TestGetNumpy.mocked_download_artifacts_numpy)
        df = mlf_obj.get_numpy('my_df')
        assert df.shape[0] == 2
        assert df.shape[1] == 3
        assert df[0][0] == 1
        assert df[0][1] == 2
        assert df[0][2] == 3
        assert df[1][0] == 11
        assert df[1][1] == 22
        assert df[1][2] == 33

    def test_operation_with_folder(self, mlf_obj, monkeypatch):
        monkeypatch.setattr(mlflow.artifacts, 'download_artifacts', TestGetNumpy.mocked_download_artifacts_numpy)
        df = mlf_obj.get_numpy('my_df', 'folder/path')
        assert df.shape[0] == 2
        assert df.shape[1] == 3
        assert df[0][0] == 1
        assert df[0][1] == 2
        assert df[0][2] == 3
        assert df[1][0] == 11
        assert df[1][1] == 22
        assert df[1][2] == 33



"""Tests metrics module"""

import pytest
import math
import numpy as np
from lpsds.metrics import sp_index, sensitivity, specificity, sp_score, ppv, npv

class TestSPIndex:
    """TEsts sp_index function"""

    def sp_func(self, v1,v2):
        """Help function for SP comparison"""
        geo_mean = math.sqrt(v1*v2)
        arit_mean = 0.5 * (v1+v2)
        return math.sqrt(geo_mean * arit_mean)

    def test_scalar(self):
        """Test if scalars are passed"""
        v1,v2 = 0.9, 0.4
        assert sp_index(v1,v2) == self.sp_func(v1,v2)


    @pytest.mark.parametrize(("vec1", "vec2"), [ 
                                                    (np.array([0.9, 0.2, 0.4]), np.array([0.89, 0.4, 0.4])), # 2 dimensionless vectors
                                                    (np.array([ [0.9, 0.2, 0.4] ]), np.array([ [0.89, 0.4, 0.4] ])), #2 row vectors
                                                    (np.array([ [0.9], [0.2], [0.4] ]), np.array([ [0.89], [0.4], [0.4] ])), #2 col vectors
                                                    (np.array([0.9, 0.2, 0.4]), np.array([ [0.89, 0.4, 0.4] ])), # mixed dimensionless $ row
                                                    (np.array([0.9, 0.2, 0.4]), np.array([ [0.89], [0.4], [0.4] ])), # 2 mixed dimwnsionless & cols
                                                    (np.array([ [0.9, 0.2, 0.4] ]), np.array([0.89, 0.4, 0.4])), # mixed row & dimensionless
                                                    (np.array([ [0.9, 0.2, 0.4] ]), np.array([ [0.89], [0.4], [0.4] ])), # mixed rown & col
                                                    (np.array([ [0.9], [0.2], [0.4] ]), np.array([0.89, 0.4, 0.4])), # mixed col & dimensionless
                                                    (np.array([ [0.9], [0.2], [0.4] ]), np.array([ [0.89, 0.4, 0.4] ])), # mixed col & row

                                                ])
    def test_arrays(self, vec1, vec2):
        """Tests array case"""
        ret = sp_index(vec1, vec2)
        for v1, v2, r in zip(vec1.flatten(), vec2.flatten(), ret):
            assert r == self.sp_func(v1,v2)



class TestSKLearnMetrics:
    """Tests sensitivity"""

    @pytest.fixture
    def y_true(self):
        return np.array([1,1,1,1,1,0,0,0,0,0])

    @pytest.fixture
    def y_pred(self):
        return np.array([0,0,1,1,1,0,1,1,1,1])

    @pytest.mark.parametrize(("metric_func", "target_val"), [
        (sensitivity, 0.6),
        (specificity, 0.2),
        (sp_score, 0.37224194364083985),
        (ppv, 0.42857142857142855),
        (npv, 0.3333333333333333),
    ])
    def test_array(self, y_true, y_pred, metric_func, target_val):
        assert metric_func(y_true, y_pred) == pytest.approx(target_val, 0.0000001)

    @pytest.mark.parametrize(("metric_func"), [
        (sensitivity),
        (specificity),
        (sp_score),
        (ppv),
        (npv),
    ])
    def test_return_float(self, y_true, y_pred, metric_func):
        """Tests whether the returned value is in float format"""
        assert type(metric_func(y_true, y_pred)) == float


"""
Handles data pre_processing related operations.
"""

import re
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


def drop_null_cols(df, threshold=0.6, inplace=False, std_nan=False, additional_nan_rep: list=None):
    """
    def drop_null_cols(df, threshold=0.6, inplace=False, std_nan=False, **kwargs)

    Remove columns which number of null samples is greater than the specified threshold.

    If std_nan is True, than the function will first standartise all nan values by using
    StandardNaN class. additional_nan_rep will be passed to the StandardNanClass constructor.
    """
    if not inplace: df = df.copy()
    if std_nan: StandardNaN(inplace=True, additional_nan_rep=additional_nan_rep).fit_transform(df)
    null_freq = df.isna().sum() / len(df)
    df.drop(columns=null_freq.loc[null_freq > threshold].index, inplace=True)
    return df

    
class BooleanEncode(BaseEstimator, TransformerMixin):
    """
    Creates an Boolean type encoding (+1, -1, 0) for categories that can be treated as such.
    Example: employed = ["Yes", "No", "Not Informed"] would become [+1, -1, 0].
    """
    def __init__(self, boolean_map: map, inplace: bool=True, dtype=None):
        """
        def __init__(self, boolean_map, inplace=True, dtype=None):

        input parameters:
          - boolean_map: a map where keys are categorical values and value is the number (+1, -1, 0)
            to be assigned to it.
        - inplace: whether to work on a copy of the data or not.
          - dtype: data type to cast results into.
        """

        if len(set(boolean_map.values())) > 3:
            raise ValueError('You cannot have more than 3 values to represent True, False and None')

        self.boolean_map = boolean_map
        self.inplace = inplace
        self.dtype = dtype
    
    def fit(self, X, y=None, **kwargs):
        """Dummy function. Nothing is done here."""
        return None
    
    def transform(self, X, **kwargs):
        """
        Apply the boolean encoding.
        """
        if not self.inplace: X = X.copy()
        X.replace(self.boolean_map, inplace=True)
        if self.dtype is not None: X = X.astype(self.dtype)
        return X
    
    def fit_transform(self, X, y=None, **kwargs):
        """Fit + transform"""
        return self.transform(X)

    def get_feature_names_out(self):
        pass


class StandardNaN(BaseEstimator, TransformerMixin):
    """
    Standartizes multiple NaN representations into an unique representation.

    Ex: "", "Not Informed", "Not Available", etc -> np.NaN
    """

    def __init__(self, additional_nan_rep: list=None, std_nan_val=np.NaN, inplace: bool=True):
        """
        def __init__(self, additional_nan_rep: dict={}, std_nan_val=np.NaN, inplace: bool=True):
        Class constructor

        Input parameters:
          - additional_nan_rep: a list with *additional* values that you wish to standartize
            with an unique NaN representation. By default, the following values are already
            included: 'na', 'nan', '', '<NA>', 'NÃO INFORMADO', 'None'.
          - std_nan_val: the unique NaN representation that will be used instead of the 
            representations passed.
          - inplace: whether to work on a copy of the data or not.
        """

        #Original NaN reps
        self.nan_rep = [
            re.compile(r'^\s*nan?\s*$', re.IGNORECASE),
            re.compile(r'^[\s\-]*$', re.IGNORECASE),
            re.compile(r'^\s*<NA>\s*$', re.IGNORECASE),
            re.compile(r'^\s*NÃO INFORMADO\s*$', re.IGNORECASE),
            re.compile(r'^\s*None\s*$', re.IGNORECASE),
        ]

        if additional_nan_rep: self.nan_rep += additional_nan_rep
        self.inplace = inplace
        self.std_nan_val = std_nan_val
    
    def fit(self, X, y=None, **kwargs):
        """Dummy function. Nothing is done here."""
        return None
    
    def transform(self, X, **kwargs):
        """Standartizes multiple NaN references"""
        if not self.inplace: X = X.copy()
        X.replace(to_replace=self.nan_rep, value=self.std_nan_val, regex=True, inplace=True)
        return X
    
    def fit_transform(self, X, y=None, **kwargs):
        """Fit + transform"""
        return self.transform(X)

    def get_feature_names_out(self):
        pass


class SmartFloatCasting(BaseEstimator, TransformerMixin):
    """
    Cast float represented as strings into a float number by taking into consideration
    specific thousand and decimal separators.
    """
    def __init__(self, mod_list=[(',', '.')], inplace: bool=True, dtype=np.float64):
        """
        def __init__(self, mod_list=[(',', '.')], inplace: bool=True, dtype=np.float64)

        Input parameters:
          - mod_list: a list of tuples in the shape (orig val, new val). Orig val may be a regexp.
                      This list will be iterated and the replacement found in each iteration will be applied.
          - inplace: whether to work on a copy of the data or not.
          - dtype: data type to cast results into.
        """
        self.mod_list = mod_list
        self.inplace = inplace
        self.dtype=dtype
    
    def fit(self, X, y=None, **kwargs):
        """Dummy function. Nothing is done here."""
        return None

    def transform(self, X, **kwargs):
        """Do the smart float casting."""
        if not self.inplace: X = X.copy()
        for c in X.columns:
            for orig, new in self.mod_list:
                X[c] = X[c].str.replace(orig, new).astype(self.dtype)
        return X
    
    def fit_transform(self, X, y=None, **kwargs):
        """Fit + transform"""
        return self.transform(X)

    def get_feature_names_out(self):
        pass



import os
import re
import yaml
import numpy as np
import pandas as pd
import tempfile
import mlflow
import lpsds.metrics
from lpsds.utils import ObjectView


class MLFlow:
    def __init__(self, run_id=None):
        active_run = mlflow.active_run()

        if run_id is None and active_run is None:
            raise ValueError('You must invoke this class either by passing a valid run id or within an started run (mlflow.start_run)')

        self.run_id = active_run.info.run_id if run_id is None else run_id
        self.run = mlflow.get_run(self.run_id)
        self.run_client = mlflow.tracking.MlflowClient()

    def log_dataframe(self, df: pd.DataFrame, var_name: str, folder: str='') -> None:
        """"
        def log_dataframe(self, df: pd.DataFrame, var_name: str, folder: str='') -> None

        Saves a pandas dataframe to MLFlow as a .parquet file.

        Input parameters:
        - df: the pandas.DataFrame object to be saved.
        - var_name: the name the dataframe will have within MLFlow.
        - folder: the path (in MLFlow) to where the dataframe will be saved.
        """

        self.log_artifact(df, var_name + '.parquet', folder, save_func=df.to_parquet, object_param_name=None, fname_param_name='path')


    def log_numpy(self, mat: np.ndarray, var_name: str, folder: str='') -> None:
        """"
        def log_numpy(self, mat: np.ndarray, var_name: str, folder: str='') -> None

        Saves a numpy matrix to MLFlow as a .np file.

        Input parameters:
        - mat: the np.ndarray object to be saved.
        - var_name: the name the dataframe will have within MLFlow.
        - folder: the path (in MLFlow) to where the dataframe will be saved.
        """

        self.log_artifact(mat, var_name + '.npy', folder, allow_pickle=False)



    def log_artifact(self, obj, var_name: str, folder: str='', save_func=np.save,
                     object_param_name='arr', fname_param_name='file', **save_func_kwargs) -> None:
        """"
        def log_artifact(self, obj, var_name: str, folder: str='', save_func=np.save,
                         object_param_name='arr', fname_param_name='file', **save_func_kwargs) -> None:

        Saves an object of any type to mlflow.

        Input parameters:
        - obj: the object you want to save to MLFlow.
        - var_name: the name the dataframe will have within MLFlow.
        - folder: the path (in MLFlow) to where the dataframe will be saved.
        - save_func: a reference to a function which is responsible to save the passed object to disk (obj.to_parquet, np.save, etc.)
        - object_param_name: the save_func parameter name used to receive obj when saving (arr in np.save, for instance).
        - fname_param_name: the save_func parameter name used to receive the path where the file will be saved (path in np.save, for instance).
        - save_func_kwargs: additional parameters to be passed to save_func.
        """

        with tempfile.TemporaryDirectory() as temp_path:
            temp_file_name = os.path.join(temp_path, var_name)
            save_func_kwargs[fname_param_name] = temp_file_name
            if object_param_name is not None:
                save_func_kwargs[object_param_name] = obj
            save_func(**save_func_kwargs)
            mlflow.log_artifact(temp_file_name, folder)



    def log_statistics(self, cv_model: dict) -> dict:
        """
        def log_statistics(cv_model)

        Log metrics obtained via cross validation to MLFlow as statistical summaries.
        I.e.: the mean value obtained across all folds and its minimum and maximum
        values so right value will be found in this range (err_min, mean, err_max)
        with 95% C.I.

        The metrics will be found automatically by looking to keys within cv_model
        object that has the pattern "test_*".

        Input:
            cv_model: the result of the sklearn cross_validate function.
        
        Returns a map where keys are the metric name and values are its mean, err min and err max
        """
        regexp = re.compile(r'test_(.+)')
        ret_map = {}
        for metric, values in cv_model.items():
            grp = regexp.match(metric)
            if grp is not None:
                mean, err_min, err_max = lpsds.metrics.bootstrap_estimate(values)
                metric_name = grp.group(1)
                ret_map[metric_name] = {
                    f'{metric_name}_err_min' : err_min,
                    f'{metric_name}_mean' : mean,
                    f'{metric_name}_err_max' : err_max
                }
                mlflow.log_metrics(ret_map[metric_name])
                for i,v in enumerate(values): mlflow.log_metric(metric_name, v, step=i)
        
        return ret_map


    def get_params(self, infer_types: bool=False) -> ObjectView:
        """
        Returns model parameters.

        Input:
          - infer_dtypes: if True, will try to infer values types, since mlflow 
                          store them as strings.
        
        Return: a map with the model parameters.
        """
        ret = ObjectView(self.run.data.params)
        if infer_types:
            for k,v in ret.items():
                ret[k] = yaml.safe_load(v)
        return ret


    def get_run_id(self) -> str:
        """
        def get_run_id(self)

        Returns the experiment´s run id.
        """
        return self.run_id


    def get_metrics(self, as_dict=False):
        """
        Collects all metrics available for a given run.
        

        Returns a pandas dataframe with all metrics.

        if as_dict is True, the method will return the metrics as
        a dict, where, If the metric is a vector, it is returned as a numpy.array.
        """
        
        #Collecting all metrics in a dataframe
        df = pd.DataFrame(columns=['metric', 'step', 'value'])
        for metric_name in self.run.data.metrics.keys():
            for m in self.run_client.get_metric_history(self.run_id, metric_name):
                df.loc[len(df)] = metric_name, m.step, m.value
        df.sort_values(['metric', 'step'], inplace=True, ignore_index=True)
        
        if not as_dict: return df
    
        #Creating a dataframe where vectorized metrics are saved as lists
        grp = df.groupby('metric').value.agg(lambda x: x.iloc[0] if len(x) == 1 else x.to_list()).to_dict()
        
        #Lists are converted to np.arrays
        for k,v in grp.items():
            if hasattr(v, '__iter__'):
                grp[k] = np.array(v)
        return ObjectView(grp)



    def get_dataframe(self, var_name: str, folder: str='') -> pd.DataFrame:
        """"
        def get_dataframe(self, var_name: str, folder: str='') -> pd.DataFrame:

        Load a pandas dataframe saved to MLFlow as a .parquet file.

        Input parameters:
        - var_name: the name the dataframe have within MLFlow.
        - folder: the path (in MLFlow) to where the dataframe will be loaded from.
        
        Return a pandas.DataFrame with the collected info.
        """

        return self.get_artifact(var_name + '.parquet', folder, pd.read_parquet)


    def get_numpy(self, var_name: str, folder: str='') -> np.ndarray:
        """"
        def get_numpy(self, var_name: str, folder: str='') -> np.ndarray:

        Load a numpy.ndarray object saved to MLFlow as a .parquet file.

        Input parameters:
        - var_name: the name the dataframe have within MLFlow.
        - folder: the path (in MLFlow) to where the dataframe will be loaded from.
        
        Return a np.ndarray with the collected info.
        """

        return self.get_artifact(var_name + '.npy', folder, np.load, allow_pickle=False)



    def get_artifact(self, var_name: str, folder: str='', load_func=np.load, **load_func_kwargs):
        """"
        def get_artifact(self, var_name: str, folder: str='', load_func=np.load, **load_func_kwargs)

        Load an artifact using the provided loading function.

        Input parameters:
        - var_name: the name the dataframe have within MLFlow (must include extensions, if existing).
        - folder: the path (in MLFlow) to where the dataframe will be loaded from.
        - load_func: the function used to load the required artifact (pd.read_parquet, np.load, etc).
        - load_func_kwargs: additional aprameters to be passed to load_func
        
        Returns whatever load_func returns.
        """

        with tempfile.TemporaryDirectory() as temp_path:
            full_path = os.path.join(folder, var_name)
            local_path = mlflow.artifacts.download_artifacts(run_id=self.run_id, artifact_path=full_path, dst_path=temp_path)
            return load_func(local_path, **load_func_kwargs)




    def get_experiment(self) -> mlflow.entities.Experiment:
        """
        def get_experiment(self) -> mlflow.entities.Experiment

        Returns an instance to the experiment that contains the class given mlflow run id.
        """
        exp_id = self.run.info.experiment_id
        return mlflow.get_experiment(exp_id)


"""Metric related tools"""

from typing import Tuple 
import numpy as np
from scipy.stats import gmean
from seaborn.algorithms import bootstrap
from sklearn.metrics import recall_score, precision_score

def bootstrap_estimate(vec: np.ndarray, ci: int=95, n_boot: int=1000, seed: int=None) -> Tuple[float, float, float]:
    """
    def bootstrap_estimate(vec: np.ndarray, ci: int=95, n_boot: int=1000, seed: int=None) -> Tuple[float, float, float]:

    Returns the aggregated result for vector vec using the same CI estimator as seaborn.

    Input:
      - vec: a numpy vector [N,]
      - ci: the confidence interval to consider.
      - n_boot: how many samplings to employ when using bootstrap for the CI interval.
      - seed: the seed value to use.

    Returns: a tuple with the following values:
      - The mean value of vec
      - The lower limit of the confidence interval
      - The upper limit of the confidence interval
    """
    def percentile_interval(data, width):
        """Return a percentile interval from data of a given width."""
        edge = (100 - width) / 2
        percentiles = edge, 100 - edge
        return np.percentile(data, percentiles)

    mean = vec.mean()
    boots = bootstrap(vec, func='mean', n_boot=n_boot, seed=seed)
    err_min, err_max = percentile_interval(boots, ci)

    return mean, err_min, err_max


def sp_index(tp: np.ndarray, tn: np.ndarray) -> np.ndarray:
  """
  def sp_index(tp: np.ndarray, tn: np.ndarray) -> np.ndarray:

  Calculates the SP index, which is given by:

  sp = \sqrt{ \sqrt{tp \times tn} \times \(\frac{tp+tn,2}\) }

  where tp is the true positive values and tn the true negative values.

  Returns: an array with the sp index calculated.
  """

  if type(tp) is np.ndarray: tp = tp.flatten()
  if type(tn) is np.ndarray: tn = tn.flatten()
  mat = np.array([tp, tn])
  return np.sqrt( gmean(mat, axis=0) * mat.mean(axis=0) ).flatten()


def sensitivity(y_true: np.ndarray, y_pred: np.ndarray) -> float:
  """
  def sensitivity(y_true: np.ndarray, y_pred: np.ndarray) -> float:

  Calculates the sensitivity score. Sklearn style.
  """
  return float(recall_score(y_true, y_pred, pos_label=1))


def specificity(y_true: np.ndarray, y_pred: np.ndarray) -> float:
  """
  def specificity(y_true: np.ndarray, y_pred: np.ndarray) -> float:

  Calculates the specificity score. Sklearn style.
  """
  return float(recall_score(y_true, y_pred, pos_label=0))


def sp_score(y_true: np.ndarray, y_pred: np.ndarray) -> float:
  """
  def sp_score(y_true: np.ndarray, y_pred: np.ndarray) -> float:

  Calculates the sp_score score. Sklearn style.
  """
  return float(sp_index(sensitivity(y_true, y_pred), specificity(y_true, y_pred)))


def ppv(y_true: np.ndarray, y_pred: np.ndarray) -> float:
  """
  def ppv(y_true: np.ndarray, y_pred: np.ndarray) -> float:

  Calculates the positive predicted value. Sklearn style.
  """
  return float(precision_score(y_true, y_pred, pos_label=1))


def npv(y_true: np.ndarray, y_pred: np.ndarray) -> float:
  """
  def npv(y_true: np.ndarray, y_pred: np.ndarray) -> float:

  Calculate the negative predicted value. Sklearn style.
  """
  return float(precision_score(y_true, y_pred, pos_label=0))



import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns


def setup(width=14., num_cols=None, num_rows=None, float_fmt='{:,.3f}'):
    mpl.rcParams['figure.figsize'] = [width, width/1.5]
    sns.set(context='talk', style='white', palette='muted')
    pd.options.display.max_columns =  num_cols
    pd.options.display.max_rows =  num_rows
    pd.options.display.float_format = float_fmt.format


setup()
